/**
 ** EstadoPartida.java
 **
 ** Copyright (C) 1992, 1999, 2007, 2011 Mariano Alvarez Fernandez
 ** [e-mail: malfer at telefonica.net]
 **
 ** This file is part of the Parchis4A program.
 **
 ** Parchis4A is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** Parchis4A is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with Parchis4A.  If not, see <http://www.gnu.org/licenses/>.
 **
 **/

package com.fgrim.parchis4a;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Random;

import android.os.Bundle;

public class EstadoPartida extends DefPartida{
//	private static final String TAG = "Parchis4A";
	private static final Random RNG = new Random();
	
	//actions
	public static final int ACC_TIRADADO = 0;
	public static final int ACC_PARADADO = 1;
	public static final int ACC_MUEVE = 3;
	public static final int ACC_PASA = 7;
	public static final int ACC_PARTIDATERMINADA = 8;
	
	public boolean jgha_salido[];
	public int jgha_terminado[];
	public int sig_pos_fin;
	public Posicion pos_ficha[][];
	public int nfcasa[];
	public int nfmeta[];
	public int ultf_selec[];
	
	public int turno;
	public int nseises;
	public int nfmovida;

	public int accion;
	public int vdado;
	public GrupoJugadas gjg;
	public int njgelegida;
	public int sigvdado;
	
	public int estDado[][];
	public int estTotDado[];
	public int estPasa[];
	public int estCome[];
	public int estComida[];
	public int estACasa[];
	
	public EstadoPartida() {
		super();
		jgha_salido = new boolean[4];
		jgha_terminado = new int[4];
		pos_ficha =  new Posicion[4][4];
		nfcasa = new int[4];
		nfmeta = new int[4];
		ultf_selec = new int[4];
		estDado = new int[4][6];
		estTotDado = new int[4];
		estPasa = new int[4];
		estCome = new int[4];
		estComida = new int[4];
		estACasa = new int[4];
	}

	public EstadoPartida(EstadoPartida auxEP) {
		this();
		for (int c=AMAR; c<=VERD; c++) {
			tipo_jugador[c] = auxEP.tipo_jugador[c];
			nivel_maquina[c] = auxEP.nivel_maquina[c];
			jgha_salido[c] = auxEP.jgha_salido[c];
			jgha_terminado[c] = auxEP.jgha_terminado[c];
			for (int nf=0; nf<4; nf++) {
				pos_ficha[c][nf] = new Posicion();
				pos_ficha[c][nf].setPosicion(auxEP.pos_ficha[c][nf]);
			}
			nfcasa[c] = auxEP.nfcasa[c];
			nfmeta[c] = auxEP.nfmeta[c];
			ultf_selec[c] = auxEP.ultf_selec[c];
			for (int d=0; d<6; d++)
				estDado[c][d] = auxEP.estDado[c][d];
			estTotDado[c] = auxEP.estTotDado[c];
			estPasa[c] = auxEP.estPasa[c];
			estCome[c] = auxEP.estCome[c];
			estComida[c] = auxEP.estComida[c];
			estACasa[c] = auxEP.estACasa[c];
		}
		sig_pos_fin = auxEP.sig_pos_fin;
		turno = auxEP.turno;
		nseises = auxEP.nseises;
		nfmovida = auxEP.nfmovida;
		fin_rapido = auxEP.fin_rapido;
		finrobot_rapido = auxEP.finrobot_rapido;
		dado_manual = auxEP.dado_manual;
		num_jg = auxEP.num_jg;
		num_jgh = auxEP.num_jgh;
		num_jgr = auxEP.num_jgr;
	}
	
    private int[] pos_ficha_to_array() {
    	int pfa[] = new int[4*4*3];
    	int k = 0;
		for (int c=AMAR; c<=VERD; c++) {
			for (int nf=0; nf<4; nf++) {
				pfa[k++] = pos_ficha[c][nf].lugar;
				pfa[k++] = pos_ficha[c][nf].pos;
				pfa[k++] = pos_ficha[c][nf].ind;
			}
		}
    	return pfa;
    }
    
    private void array_to_pos_ficha(int pfa[]) {
    	int k = 0;
		for (int c=AMAR; c<=VERD; c++) {
			for (int nf=0; nf<4; nf++) {
				pos_ficha[c][nf] = new Posicion();
				pos_ficha[c][nf].lugar = pfa[k++];
				pos_ficha[c][nf].pos = pfa[k++];
				pos_ficha[c][nf].ind = pfa[k++];
			}
		}
    }
    
	/* Save game state in a bundle*/
    public Bundle saveState() {
        Bundle map = new Bundle();

        map.putIntArray("tj", tipo_jugador);
        map.putIntArray("nm", nivel_maquina);
        map.putBooleanArray("hs", jgha_salido);
        map.putIntArray("ht", jgha_terminado);
        map.putInt("spfi", sig_pos_fin);
        map.putIntArray("pf", this.pos_ficha_to_array());
        map.putIntArray("nfcs", nfcasa);
        map.putIntArray("nfmt", nfmeta);
        map.putIntArray("ulfs", ultf_selec);
        map.putBoolean("frp", fin_rapido);
        map.putBoolean("frrp", finrobot_rapido);
        map.putBoolean("dm", dado_manual);
        map.putInt("njg", num_jg);
        map.putInt("njgh", num_jgh);
        map.putInt("njgr", num_jgr);
        map.putInt("tur", turno);
        map.putInt("n6", nseises);
        map.putInt("nfmv", nfmovida);
        map.putInt("acc", accion);
        map.putInt("vd", vdado);
        map.putInt("njge", njgelegida);
        map.putInt("sigv", sigvdado);
        map.putIntArray("estd1", estDado[0]);
        map.putIntArray("estd2", estDado[1]);
        map.putIntArray("estd3", estDado[2]);
        map.putIntArray("estd4", estDado[3]);
        map.putIntArray("esttd", estTotDado);
        map.putIntArray("estp", estPasa);
        map.putIntArray("estc", estCome);
        map.putIntArray("estm", estComida);
        map.putIntArray("estac", estACasa);

        return map;
    }

    public void restoreState(Bundle icicle) {
    	tipo_jugador = icicle.getIntArray("tj");
    	nivel_maquina = icicle.getIntArray("nm");
    	jgha_salido = icicle.getBooleanArray("hs");
    	jgha_terminado = icicle.getIntArray("ht");
    	sig_pos_fin = icicle.getInt("spfi");
    	this.array_to_pos_ficha(icicle.getIntArray("pf"));
    	nfcasa = icicle.getIntArray("nfcs");
    	nfmeta = icicle.getIntArray("nfmt");
    	ultf_selec = icicle.getIntArray("ulfs");
    	fin_rapido = icicle.getBoolean("frp");
    	finrobot_rapido = icicle.getBoolean("frrp");
    	dado_manual = icicle.getBoolean("dm");
    	num_jg = icicle.getInt("njg");
    	num_jgh = icicle.getInt("njgh");
    	num_jgr = icicle.getInt("njgr");
    	turno = icicle.getInt("tur");
    	nseises = icicle.getInt("n6");
    	nfmovida = icicle.getInt("nfmv");
    	accion = icicle.getInt("acc");
    	vdado = icicle.getInt("vd");
    	njgelegida = icicle.getInt("njge");
    	sigvdado = icicle.getInt("sigv");
    	estDado[0] = icicle.getIntArray("estd1");
    	estDado[1] = icicle.getIntArray("estd2");
    	estDado[2] = icicle.getIntArray("estd3");
    	estDado[3] = icicle.getIntArray("estd4");
    	estTotDado = icicle.getIntArray("esttd");
    	estPasa = icicle.getIntArray("estp");
    	estCome = icicle.getIntArray("estc");
    	estComida = icicle.getIntArray("estm");
    	estACasa = icicle.getIntArray("estac");
    }

    /* Save game state in a file*/
    public void saveState(ObjectOutputStream oout) throws IOException {
        oout.writeObject(tipo_jugador);
        oout.writeObject(nivel_maquina);
        oout.writeObject(jgha_salido);
        oout.writeObject(jgha_terminado);
        oout.writeInt(sig_pos_fin);
        oout.writeObject(this.pos_ficha_to_array());
        oout.writeObject(nfcasa);
        oout.writeObject(nfmeta);
        oout.writeObject(ultf_selec);
        oout.writeBoolean(fin_rapido);
        oout.writeBoolean(finrobot_rapido);
        oout.writeBoolean(dado_manual);
        oout.writeInt(num_jg);
        oout.writeInt(num_jgh);
        oout.writeInt(num_jgr);
        oout.writeInt(turno);
        oout.writeInt(nseises);
        oout.writeInt(nfmovida);
        oout.writeInt(accion);
        oout.writeInt(vdado);
        oout.writeInt(njgelegida);
        oout.writeInt(sigvdado);
        oout.writeObject(estDado[0]);
        oout.writeObject(estDado[1]);
        oout.writeObject(estDado[2]);
        oout.writeObject(estDado[3]);
        oout.writeObject(estTotDado);
        oout.writeObject(estPasa);
        oout.writeObject(estCome);
        oout.writeObject(estComida);
        oout.writeObject(estACasa);
    }

    public void restoreState(ObjectInputStream oin) throws IOException, ClassNotFoundException {
    	tipo_jugador = (int[])oin.readObject();
    	nivel_maquina = (int[])oin.readObject();
    	jgha_salido = (boolean[])oin.readObject();
    	jgha_terminado = (int[])oin.readObject();
    	sig_pos_fin = oin.readInt();
    	this.array_to_pos_ficha((int[])oin.readObject());
    	nfcasa = (int[])oin.readObject();
    	nfmeta = (int[])oin.readObject();
    	ultf_selec = (int[])oin.readObject();
    	fin_rapido = oin.readBoolean();
    	finrobot_rapido = oin.readBoolean();
    	dado_manual = oin.readBoolean();
    	num_jg = oin.readInt();
    	num_jgh = oin.readInt();
    	num_jgr = oin.readInt();
    	turno = oin.readInt();
    	nseises = oin.readInt();
    	nfmovida = oin.readInt();
    	accion = oin.readInt();
    	vdado = oin.readInt();
    	njgelegida = oin.readInt();
    	sigvdado = oin.readInt();
    	estDado[0] = (int[])oin.readObject();
    	estDado[1] = (int[])oin.readObject();
    	estDado[2] = (int[])oin.readObject();
    	estDado[3] = (int[])oin.readObject();
    	estTotDado = (int[])oin.readObject();
    	estPasa = (int[])oin.readObject();
    	estCome = (int[])oin.readObject();
    	estComida = (int[])oin.readObject();
    	estACasa = (int[])oin.readObject();
    }

	public void iniciaPartida(DefPartida defP) {
		this.set(defP);
		for (int c=AMAR; c<=VERD; c++) {
			jgha_salido[c] = false;
			jgha_terminado[c] = -1;
			for (int nf=0; nf<4; nf++) {
				pos_ficha[c][nf] = new Posicion(Posicion.CASA, 0, nf);
			}
			nfcasa[c] = 4;
			nfmeta[c] = 0;
			ultf_selec[c] = -1;
			for (int d=0; d<6; d++)
				estDado[c][d] = 0;
			estTotDado[c] = 0;
			estPasa[c] = 0;
			estCome[c] = 0;
			estComida[c] = 0;
			estACasa[c] = 0;
		}
		
		sig_pos_fin = 0;
		turno = VERD; // so next player is AMAR!
        nseises = 0;
        
		accion = ACC_TIRADADO;
		sigvdado = 0;
        njgelegida = -1;
		this.avanzaTurno();
	}
	
    private void sumaTurno() {
		if (turno == VERD)
			turno = AMAR;
		else
			turno++;
    }
    
	public void avanzaTurno() {
		boolean encontrado = false;

		nseises = 0;
		nfmovida = -1;
		sumaTurno();
		
		for (int i=0; i<4; i++) {
			if (tipo_jugador[turno] != DefPartida.NOJUEGA && jgha_terminado[turno] < 0) {
				encontrado = true;
				break;
			} else
				sumaTurno();
		}
		
		if (!encontrado) {
			accion = ACC_PARTIDATERMINADA;
			return;
		}
		
		if (fin_rapido && num_jg > 1) {
		    if (num_jg == sig_pos_fin + 1) {
		    	jgha_terminado[turno] = sig_pos_fin++;
				accion = ACC_PARTIDATERMINADA;
				return;
		    }
		}
		
		if (finrobot_rapido && num_jgr > 0) {
			int njght = 0;
			int njgrt = 0;
			int ljgrst[] = new int[4];
			int njgrst = 0;
			for (int c=AMAR; c<=VERD; c++) {
				if (jgha_terminado[c] >= 0) {
					if (tipo_jugador[c] == DefPartida.PERSONA)
						njght++;
					else if (tipo_jugador[c] == DefPartida.MAQUINA)
						njgrt++;
				} else if (tipo_jugador[c] == DefPartida.MAQUINA) {
					ljgrst[njgrst] = c;
					njgrst++;
				}
			}
			if ((num_jgh > 0 && njght == num_jgh)) { // han terminado todos los humanos
//				(num_jgh == 0 && njgrt > 0)) {       // o solo robots y ya ha entrado uno
				if (njgrst == 1) { // solo queda un robot
			    	jgha_terminado[ljgrst[0]] = sig_pos_fin++;
					accion = ACC_PARTIDATERMINADA;
					return;
				} else if (njgrst > 1) { // quedan 2 o 3
					int dist[] = new int[3];
					for (int k=0; k<njgrst; k++) {
						dist[k] = calculaDistancia(ljgrst[k]);
//						Log.d(TAG, "Dist " + k + " " +  ljgrst[k] + " " + dist[k]);
					}
					for (int i=0; i<njgrst; i++) {
						int minimo = 1000;
						int pos_minimo = 0;
						for (int k=0; k<njgrst; k++) {
							if (dist[k] < minimo) {
								minimo = dist[k];
								pos_minimo = k;
							}
						}
						dist[pos_minimo] = 2000;
				    	jgha_terminado[ljgrst[pos_minimo]] = sig_pos_fin++;
//				    	Log.d(TAG, "Posi " + pos_minimo + " " + ljgrst[pos_minimo] + " " + sig_pos_fin);
					}
					accion = ACC_PARTIDATERMINADA;
					return;
				}
			}
		}
	}
	
	public void iniciaAccion(Tablero mTbl, int vdadoManual) {
		if (accion == ACC_PARADADO) {
            if (vdadoManual > 0)
            	vdado = vdadoManual;
            else
			    vdado = RNG.nextInt(6) + 1;
			nseises = (vdado == 6) ? nseises+1 : 0;
	    	estDado[turno][vdado-1]++;
	    	estTotDado[turno]++;
			if (vdado == 6 && nfcasa[turno] == 0) vdado = 7; // hace 6 a 7
		} else if (accion == ACC_MUEVE) {
//			gjg.jg[njgelegida].impLog();
	    	if (gjg.jg[njgelegida].come) {
	    		estCome[gjg.jg[njgelegida].cjugador]++;
	    		estComida[gjg.jg[njgelegida].cjugdcom]++;
	    	}
	    	if (gjg.jg[njgelegida].acasa) {
	    		estACasa[gjg.jg[njgelegida].cjugador]++;
	    	}
			sigvdado = mTbl.realizaJugada(gjg.jg[njgelegida]);
			if (gjg.njugadas > 1) {
				ultf_selec[gjg.jg[njgelegida].cjugador] = gjg.jg[njgelegida].nficha;
			}
            njgelegida = -1;
		} else if (accion == ACC_PASA) {
			estPasa[turno]++;
			sigvdado = 0;
		}
	}

	private int jugadaPorDefecto(int nj) {
		int jeleg = 0;
		
		if (nj < 2) return jeleg;
		int color = gjg.jg[0].cjugador;
		if (ultf_selec[color] < 0) return jeleg;
		for (int i=0; i<nj; i++) {
			if (gjg.jg[i].nficha == ultf_selec[color]) {
				jeleg = i;
				break;
			}
		}
		return jeleg;
	}
	
	public void terminaAccion(Tablero mTbl) {
		if (accion == ACC_TIRADADO) {
			accion = ACC_PARADADO;
		} else if (accion == ACC_PARADADO) {
            gjg = new GrupoJugadas();
            int n = gjg.genera(mTbl, turno, vdado);
            if (n > 0) {
                njgelegida = jugadaPorDefecto(n);
                if (tipo_jugador[turno] == DefPartida.MAQUINA)
                	eligeJugadaMaquina(mTbl);
            	accion = ACC_MUEVE;
            } else {
                njgelegida = -1;
            	accion = ACC_PASA;
            }
		} else if (accion == ACC_PASA) {
		    accion = ACC_TIRADADO;
		    if (sigvdado == -1 || nseises == 0 || nseises >2)
  		        avanzaTurno();
		} else if (accion == ACC_MUEVE) {
			if (sigvdado == 10 || sigvdado == 20) {
				vdado = sigvdado;
	            gjg = new GrupoJugadas();
	            int n = gjg.genera(mTbl, turno, vdado);
	            if (n > 0) {
	                njgelegida = jugadaPorDefecto(n);
	                if (tipo_jugador[turno] == DefPartida.MAQUINA)
	                	eligeJugadaMaquina(mTbl);
	            	accion = ACC_MUEVE;
	            } else {
	                njgelegida = -1;
	            	accion = ACC_PASA;
	            }
			} else {
  		        accion = ACC_TIRADADO;
  		        if (sigvdado == -1 || nseises == 0 || nseises >2)
		            avanzaTurno();
			}
	    }
	}
	
	public void eligeJugadaMaquina(Tablero mTbl) {
		int i, v, valor = -32000;
		
        if (gjg.njugadas < 2) return;
        
		gjg.completaInfoAnl(mTbl);
		for (i=0; i<gjg.njugadas; i++) {
			v = gjg.jg[i].calculaPeso(nivel_maquina[turno]);
			if (v > valor) {
				valor = v;
				njgelegida = i;
			}
		}
	}
	
	public boolean haComenzado() {
		for (int c=AMAR; c<=VERD; c++) {
	        if (nfcasa[c] < 4) return true;
		}
		return false;
	}
	
	public boolean haTerminado() {
		for (int c=AMAR; c<=VERD; c++) {
		    if (tipo_jugador[c] != DefPartida.NOJUEGA && jgha_terminado[c] < 0) return false;
		}
		return true;
	}

	public int calculaDistancia(int color) {
		int distancia = 0;
		for (int nf=0; nf<4; nf++) {
			if (pos_ficha[color][nf].lugar == Posicion.META)
				continue;
			else if (pos_ficha[color][nf].lugar == Posicion.LLEGADA)
				distancia += 7 - pos_ficha[color][nf].pos;
			else if (pos_ficha[color][nf].lugar == Posicion.CORREDOR) {
				int pos = pos_ficha[color][nf].pos;
		    	if (pos < Jugada.cent[color]) pos += 68;
		    	pos = pos - (17 * color);
		    	distancia += 7 + 68 - pos; 
			} else // CASA
				distancia += 7 + 68;
		}
		return distancia;
	}

	public String impEstadisticas() {
		String s;
		
		s = "";
		for (int c=AMAR; c<=VERD; c++) {
			for (int ind=0; ind<6; ind++) {
				s += (ind+1) + ":" + estDado[c][ind] + " "; 
			}
			s += "P:" + estPasa[c] + " ";
			s += "C:" + estCome[c] + " ";
			s += "Cmd:" + estComida[c] + " ";
			s += "AC:" + estACasa[c] + "\n";
		}
		return s;
	}
}
