/**
 ** Jugada.java
 **
 ** Copyright (C) 1992, 1999, 2007, 2011 Mariano Alvarez Fernandez
 ** [e-mail: malfer at telefonica.net]
 **
 ** This file is part of the Parchis4A program.
 **
 ** Parchis4A is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** Parchis4A is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with Parchis4A.  If not, see <http://www.gnu.org/licenses/>.
 **
 **/

package com.fgrim.parchis4a;

import java.util.Random;

import android.util.Log;

public class Jugada {
	private static final String TAG = "Parchis4A";

	private static final Random RNG = new Random();
    
    public static final int cent[] = { 4,21,38,55 };
    public static final int csal[] = { 67,16,33,50 };
    
	public int cjugador;
	public int nficha;
	public int vdado;
	public int cjugdcom;
	public int nfichcom;
	public Posicion origen;
	public Posicion destino;
	
	// Analisis
	public boolean entra;   /* entra en el corredor */
	public boolean sale;    /* sale del corredor */
	public boolean meta;    /* entra en la meta */
	public boolean acasa;   /* a casa por tres seises */
	public boolean come;    /* come ficha */
	public boolean hpte;    /* hace puente */
	public boolean rpte;    /* rompe puente */
	public boolean dseg;    /* deja seguro */
	public boolean aseg;    /* va a seguro */
	public int nficom;  /* n. de fichas comibles en la posicion inicial */
	public int nfipel;  /* n. de fichas peligrosas en la posicion inicial */
	public int nffcom;  /* n. de fichas comibles en la posicion final */
	public int nffpel;  /* n. de fichas peligrosas en la posicion final */
	public int ncfin;   /* n. de casillas que faltan para llegar */
	
	public Jugada() {
		origen = new Posicion();
		destino = new Posicion();
		entra = false;
		sale = false;
		meta = false;
		acasa = false;
		come = false;
		hpte = false;
		rpte = false;
		dseg = false;
		aseg = false;
		nficom = 0;
		nfipel = 0;
		nffcom = 0;
		nffpel = 0;
		ncfin = 0;
	}
	
	public int genera(Tablero tbl, int ncjugador, int nnficha, int nvdado) {
		cjugador =  ncjugador;
		nficha = nnficha;
		vdado = nvdado;
		cjugdcom = Tablero.VACIO;
		nfichcom = 0;
		origen = tbl.mEP.pos_ficha[cjugador][nficha];
		
		if (origen.lugar == Posicion.META) return 0;
		
		/* siempre se supone que las casillas se rellenan por orden, no
	       puede haber una ficha en ind=1 si no la hay en ind=0 */
	    /* solo sale de casa si se saco un 5 y no hay dos fichas propias en
	       la salida, si hay 2 fichas y una no es propia, ocupa su puesto */
		
		if (origen.lugar == Posicion.CASA) {
			if (vdado != 5) return 0;
			entra = true;
			Tablero.Casillac cac = tbl.corredor[cent[cjugador]];
			if (cac.cf[0] == cjugador && cac.cf[1] == cjugador) return 0;
			destino.lugar = Posicion.CORREDOR;
			destino.pos = cent[cjugador];
			destino.ind = 0;
			if (cac.cf[0] == Tablero.VACIO) return 1;
			if (!tbl.mEP.jgha_salido[cjugador]) {
				/* como sacara dos fichas se come al menos la de ind 0
				   y seguro que no es de las suyas */
				come = true;
	        	cjugdcom = cac.cf[0];
	        	nfichcom = cac.nf[0];
	        	return 1;
			}
			if (cac.cf[1] == Tablero.VACIO) {
				destino.ind = 1;
				return 1;
			}
			come = true;
            if (cac.cf[1] != cjugador) {
            	cjugdcom = cac.cf[1];
            	nfichcom = cac.nf[1];
            	destino.ind = 1;
            	return 1;
            }
        	cjugdcom = cac.cf[0];
        	nfichcom = cac.nf[0];
        	return 1;
		}
		
		if (origen.lugar == Posicion.CORREDOR) {
			/* Mira si rompe puente */
			if (tbl.corredor[origen.pos].cf[0] == tbl.corredor[origen.pos].cf[1])
				rpte = true;
			/* Mira si deja seguro */
			int tipo = tbl.corredor[origen.pos].tipo;
			if (tipo == Tablero.SEGURO || tipo == Tablero.SALIDA)
				dseg = true;
			if (tipo >= Tablero.ENTRADA) {
			    int i = tipo - Tablero.ENTRADA;
			    if (i == cjugador ||
				    tbl.mEP.tipo_jugador[i] == DefPartida.NOJUEGA ||
				    tbl.mEP.nfcasa[i] == 0 ||
				    (tbl.mEP.nfcasa[i] < 4 &&
				     tbl.corredor[origen.pos].cf[1] == Tablero.VACIO))
				    dseg = true;
			}
			/* Mueve */
			destino.lugar = Posicion.CORREDOR;
			destino.pos = origen.pos;
			boolean en_llegada = false;
			for (int i=0; i<nvdado; i++) {
				if (destino.pos == csal[cjugador]) {
					sale = true;
					destino.lugar = Posicion.LLEGADA;
					destino.pos = -1; /* posicion ficticia */
					nvdado -= i;
					en_llegada = true;
					break;
				}
				destino.pos++;
				if (destino.pos > 67) destino.pos = 0;
				if (tbl.corredor[destino.pos].cf[1] != Tablero.VACIO) {
					/* hay una barrera, pero puede ser de fichas distintas en seguro */
					/* de todas formas, si es la posicion final, no vale */
					if (i == nvdado - 1) return 0;
					if (tbl.corredor[destino.pos].cf[0] ==
						tbl.corredor[destino.pos].cf[1]) return 0;
				}
			}
			if (!en_llegada) {
				/* estamos en la posicion final, ver si estamos solos o si se come */
			    /* o si formamos barrera, o si estamos en seguro y no se come */
				tipo = tbl.corredor[destino.pos].tipo;
				if (tipo == Tablero.SEGURO || tipo == Tablero.SALIDA)
					aseg = true;
				if (tipo >= Tablero.ENTRADA) {
				    int i = tipo - Tablero.ENTRADA;
				    if (i == cjugador ||
					    tbl.mEP.tipo_jugador[i] == DefPartida.NOJUEGA ||
					    tbl.mEP.nfcasa[i] == 0 ||
					    (tbl.mEP.nfcasa[i] < 4 &&
					     tbl.corredor[destino.pos].cf[0] == Tablero.VACIO))
					    aseg = true;
				}
				if (tbl.corredor[destino.pos].cf[0] == Tablero.VACIO) {
					destino.ind = 0;
					return 1; /* ok, sola */
				}
				if (tbl.corredor[destino.pos].cf[0] == cjugador) {
					destino.ind = 1;
					hpte = true;
					return 1; /* ok, hace puente */
				}
				if (tbl.corredor[destino.pos].tipo != Tablero.NORMAL) {
					destino.ind = 1;
					return 1; /* ok, en seguro con otra */
				}
				destino.ind = 0;
				come = true;
				cjugdcom = tbl.corredor[destino.pos].cf[0];
				nfichcom = tbl.corredor[destino.pos].nf[0];
				return 1; /* ok, come */
			}
		}
		
		if (origen.lugar == Posicion.LLEGADA) {
			destino.lugar = Posicion.LLEGADA;
			destino.pos = origen.pos;
		}
		/* si estamos aquí es porque el origen es LLEGADA o CORREDOR */
		/* con el dado descontado */
		if (destino.pos + nvdado > 7) return 0;
		/* mirar si hay algún puente en el camino (gracias M.Angeles) */
		for (int ist=1; ist < nvdado; ist++) {
			if (tbl.llegada[cjugador][destino.pos+ist].cf[1] != Tablero.VACIO)
				return 0; /* hay puente */
		}
		/* camino libre */
		if (destino.pos + nvdado == 7) {
			destino.lugar = Posicion.META;
			meta = true;
			int i;
			for (i=0; tbl.meta[cjugador].cf[i] != Tablero.VACIO; i++);
			destino.ind = i;
			return 1; /* ok, en la meta */
		}
		destino.pos += nvdado;
		if (tbl.llegada[cjugador][destino.pos].cf[1] != Tablero.VACIO)
			return 0;
		destino.ind = 0;
		if (tbl.llegada[cjugador][destino.pos].cf[0] != Tablero.VACIO)
			destino.ind = 1;
		return 1;
	}

	public int genACasa(Tablero tbl, int ncjugador, int nnficha, int nvdado) {
		cjugador =  ncjugador;
		nficha = nnficha;
		vdado = nvdado;
		cjugdcom = Tablero.VACIO;
		nfichcom = 0;
		origen = tbl.mEP.pos_ficha[cjugador][nficha];
		
		if (origen.lugar != Posicion.CORREDOR) return 0; // pero debe ser
		
        acasa = true;
        destino.lugar = Posicion.CASA;
		destino.pos = 0;
    	int i;
    	for (i=0; tbl.casa[cjugador].cf[i] != Tablero.VACIO; i++);
		destino.ind = i;
		return 1;
	}

	public int calculaPeso(int nivel) {
       int peso = 0;

	   peso += RNG.nextInt(10);  /* un poco de incertidumbre */
       if (nivel == 0) return peso;  /* nivel 0 juega a lo loco */
	   if( sale ) peso += 100;
	   if( meta ) peso += 200;
	   if( come ) peso += 500;
	   if( hpte ) peso += 15;
	   if( dseg ) peso -= 20;
	   if( aseg ) peso += 20;
	   if( dseg && aseg ) peso += 30;
       if (nivel == 1) return peso;  /* nivel 1 no analiza */
	   peso += (nffcom - nficom) * 55;
	   peso += (nfipel - nffpel) * 50;
	   peso += ncfin / 5;
	   return peso;
	}
	
	public void impLog() {
    	String s;
    	s = cjugador + " " + nficha + " " + cjugdcom + " " + nfichcom + " " + vdado;
    	s += " * " + origen.lugar + " " + origen.pos + " " + origen.ind;
    	s += " * " + destino.lugar + " " + destino.pos + " " + destino.ind;
    	s += " + " + entra + " " + come + " " + acasa;
    	Log.d(TAG, s);
		
	}
}
