/**
 ** Tablero.java
 **
 ** Copyright (C) 1992, 1999, 2007, 2011 Mariano Alvarez Fernandez
 ** [e-mail: malfer at telefonica.net]
 **
 ** This file is part of the Parchis4A program.
 **
 ** Parchis4A is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** Parchis4A is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with Parchis4A.  If not, see <http://www.gnu.org/licenses/>.
 **
 **/

package com.fgrim.parchis4a;

public class Tablero {
    // pawn colors
	public static final int AMAR = 0;
	public static final int AZUL = 1;
	public static final int ROJO = 2;
	public static final int VERD = 3;
	public static final int VACIO = -1;

	// corredor casilla types
	public static final int NORMAL = 0;
	public static final int SEGURO = 1;
	public static final int SALIDA = 2;
	public static final int ENTRADA = 3;
	public static final int ENTRADA_AMAR = 3;
	public static final int ENTRADA_AZUL = 4;
	public static final int ENTRADA_ROJO = 5;
	public static final int ENTRADA_VERD = 6;
	
    public Casilla4 casa[];
    public Casillac corredor[];
    public Casilla2 llegada[][];
    public Casilla4 meta[];

    public EstadoPartida mEP = null;

    public Tablero(){
    	casa = new Casilla4[4];
    	corredor = new Casillac[68];
    	llegada = new Casilla2[4][7];
    	meta = new Casilla4[4];
    	
    	for (int i=0; i<68; i++) {
    		corredor[i] = new Casillac();
    		corredor[i].tipo = NORMAL;
    		corredor[i].cf[0] = VACIO;
    		corredor[i].cf[1] = VACIO;
    	}
    	corredor[4].tipo   = ENTRADA_AMAR;
    	corredor[11].tipo  = SEGURO;
    	corredor[16].tipo  = SALIDA;
    	corredor[21].tipo  = ENTRADA_AZUL;
    	corredor[28].tipo  = SEGURO;
    	corredor[33].tipo  = SALIDA;
    	corredor[38].tipo  = ENTRADA_ROJO;
    	corredor[45].tipo  = SEGURO;
    	corredor[50].tipo  = SALIDA;
    	corredor[55].tipo  = ENTRADA_VERD;
    	corredor[62].tipo  = SEGURO;
    	corredor[67].tipo  = SALIDA;
    	
   	    for (int c=AMAR; c<=VERD; c++) {
   	    	casa[c] = new Casilla4();
    		for (int j=0; j<4; j++) casa[c].cf[j] = VACIO;
   	    	meta[c] = new Casilla4();
    		for (int j=0; j<4; j++) meta[c].cf[j] = VACIO;
    		for (int j=0; j<7; j++) {
    			llegada[c][j] = new Casilla2();
    			llegada[c][j].cf[0] = VACIO;
    			llegada[c][j].cf[1] = VACIO;
    		}
   	    }
    }

    public void setEstadoPartida(EstadoPartida ep) {
    	mEP = ep;
    	for (int c=AMAR; c<=VERD; c++) {
    		if (mEP.tipo_jugador[c] == DefPartida.NOJUEGA) continue;
			for (int nf=0; nf<4; nf++) {
	        	int pos = mEP.pos_ficha[c][nf].pos;
	        	int ind = mEP.pos_ficha[c][nf].ind;
                if (mEP.pos_ficha[c][nf].lugar == Posicion.CASA) {
                	casa[c].cf[ind] = c;
                	casa[c].nf[ind] = nf;
                } else if (mEP.pos_ficha[c][nf].lugar == Posicion.CORREDOR) {
                	corredor[pos].cf[ind] = c;
                	corredor[pos].nf[ind] = nf;
                } else if (mEP.pos_ficha[c][nf].lugar == Posicion.LLEGADA) {
                	llegada[c][pos].cf[ind] = c;
                	llegada[c][pos].nf[ind] = nf;
                }else if (mEP.pos_ficha[c][nf].lugar == Posicion.META) {
                	meta[c].cf[ind] = c;
                	meta[c].nf[ind] = nf;
                }
			}
    	}
    }

    public int realizaJugada(Jugada jg) {
    	mEP.nfmovida = jg.nficha;
    	/* caso especial a casa por tres seises */
    	if (jg.acasa) {
    		if (jg.origen.lugar == Posicion.CORREDOR &&
    			jg.destino.lugar == Posicion.CASA) { // por si acaso
    			this.vaciaCasillaCorredor(jg.origen.pos, jg.origen.ind);
    	    	/* actualiza resumen */
    	    	mEP.nfcasa[jg.cjugador]++;
    	    	/* coloca la nueva posicion en los datos de la ficha */
    	    	mEP.pos_ficha[jg.cjugador][jg.nficha].setPosicion(jg.destino);
    	    	/* coloca la ficha en la nueva posicion */
    	    	casa[jg.cjugador].cf[jg.destino.ind] = jg.cjugador;
    	    	casa[jg.cjugador].nf[jg.destino.ind] = jg.nficha;
    		}
    		return 0;
    	}
    	/* actualiza resumen */
    	if (jg.entra) mEP.nfcasa[jg.cjugador]--;
    	if (jg.meta) mEP.nfmeta[jg.cjugador]++;
    	/* libera la posicion inicial */
    	if (jg.origen.lugar == Posicion.CASA)
    		this.vaciaCasillaCasa(jg.cjugador, jg.origen.ind);
    	else if (jg.origen.lugar == Posicion.CORREDOR)
    		this.vaciaCasillaCorredor(jg.origen.pos, jg.origen.ind);
    	else /* no puede ser la META, luego es la LLEGADA */
    		this.vaciaCasillaLLegada(jg.cjugador, jg.origen.pos, jg.origen.ind);
    	
    	/* coloca la nueva posicion en los datos de la ficha */
    	mEP.pos_ficha[jg.cjugador][jg.nficha].setPosicion(jg.destino);
    	
    	/* coloca la ficha en la nueva posicion y mira si se come ficha */
    	boolean ficha_comida = false;
    	boolean en_meta = false;
    	boolean jfin = false;
    	if (jg.destino.lugar == Posicion.CORREDOR) {
    		ficha_comida = this.aCasa(jg.destino.pos, jg.destino.ind);
    		corredor[jg.destino.pos].cf[jg.destino.ind] = jg.cjugador;
    		corredor[jg.destino.pos].nf[jg.destino.ind] = jg.nficha;
    	} else if (jg.destino.lugar == Posicion.LLEGADA) {
    		llegada[jg.cjugador][jg.destino.pos].cf[jg.destino.ind] = jg.cjugador;
    		llegada[jg.cjugador][jg.destino.pos].nf[jg.destino.ind] = jg.nficha;
    	} else { /* debe ser la meta */
    		meta[jg.cjugador].cf[jg.destino.ind] = jg.cjugador;
    		meta[jg.cjugador].nf[jg.destino.ind] = jg.nficha;
    		if (mEP.nfmeta[jg.cjugador] < 4)
    			en_meta = true;
    		else {
    			jfin = true;
    			mEP.jgha_terminado[jg.cjugador] = mEP.sig_pos_fin++;
    		}
    	}
    	
    	/* si es el primer 5 hay que sacar dos fichas */
    	if (jg.entra && !mEP.jgha_salido[jg.cjugador]) {
    		mEP.nfcasa[jg.cjugador]--;
    		Posicion paux = new Posicion();
    		paux.setPosicion(jg.destino);
    		paux.ind = (jg.destino.ind == 0) ? 1 : 0;
    		this.vaciaCasillaCasa(jg.cjugador, 0);
    		mEP.pos_ficha[jg.cjugador][1].setPosicion(paux);
    		ficha_comida = this.aCasa(paux.pos, paux.ind);
    		corredor[paux.pos].cf[paux.ind] = jg.cjugador;
    		corredor[paux.pos].nf[paux.ind] = 1;
    		mEP.jgha_salido[jg.cjugador] = true;
    	}
    	
    	/* devuelve valor de dado de la jugada */
    	if (ficha_comida) return 20;
    	if (en_meta) return 10;
    	if (jfin) return -1; /* indica que ya han llegado todas sus fichas */
    	return 0;
    }

    public void vaciaCasillaCasa(int cjugador, int ncasilla) {
    	for (int i=ncasilla; i<3; i++) {
    		casa[cjugador].cf[i] = casa[cjugador].cf[i+1];
    		casa[cjugador].nf[i] = casa[cjugador].nf[i+1];
    		if (casa[cjugador].cf[i] != VACIO)
    			mEP.pos_ficha[cjugador][casa[cjugador].nf[i]].ind--;
    	}
    	casa[cjugador].cf[3] = VACIO;
    }

    public void vaciaCasillaCorredor(int npos, int ncasilla) {
    	if (ncasilla == 0) {
    		corredor[npos].cf[0] = corredor[npos].cf[1];
    		corredor[npos].nf[0] = corredor[npos].nf[1];
    		int cj = corredor[npos].cf[0];
    		if (cj != VACIO)
    			mEP.pos_ficha[cj][corredor[npos].nf[0]].ind--;
    	}
    	corredor[npos].cf[1] = VACIO;
    }

    public void vaciaCasillaLLegada(int cjugador, int npos, int ncasilla) {
    	if (ncasilla == 0) {
    		llegada[cjugador][npos].cf[0] = llegada[cjugador][npos].cf[1];
    		llegada[cjugador][npos].nf[0] = llegada[cjugador][npos].nf[1];
    		if (llegada[cjugador][npos].cf[0] != VACIO)
    			mEP.pos_ficha[cjugador][llegada[cjugador][npos].nf[0]].ind--;
    	}
    	llegada[cjugador][npos].cf[1] = VACIO;
    }

    public boolean aCasa(int npos, int ncasilla) {
    	int cf = corredor[npos].cf[ncasilla];
    	int nf = corredor[npos].nf[ncasilla];
    	if (cf == VACIO) return false;
    	
    	/* actualiza resumen */
    	mEP.nfcasa[cf]++;
    	
    	int ind;
    	for (ind=0; casa[cf].cf[ind] != VACIO; ind++);
    	
    	casa[cf].cf[ind] = cf;
    	casa[cf].nf[ind] = nf;
    	mEP.pos_ficha[cf][nf].setPosicion(Posicion.CASA, 0, ind);
    	
    	return true;
    }

	public static class Casillac {
    	int tipo;
    	int cf[];
    	int nf[];
    	
    	public Casillac() {
    		cf = new int[2];
    		nf = new int[2];
    	}
    }

    public static class Casilla2 {
   	    int cf[];
    	int nf[];
    	
    	public Casilla2() {
    		cf = new int[2];
    		nf = new int[2];
    	}
    }

    public static class Casilla4 {
    	int cf[];
    	int nf[];
    	
    	public Casilla4() {
    		cf = new int[4];
    		nf = new int[4];
    	}
    }
}
